import { LaravelPassportClientOptions } from '../interfaces/LaravelPassportClientOptions';
export declare class LaravelPassportClient implements LaravelPassportClientOptions {
    domain: string;
    client_id: string;
    redirect_uri: string;
    private _token?;
    private _oauthPrefix?;
    get oauthPrefix(): string;
    set oauthPrefix(value: string);
    private _scope?;
    get scope(): string;
    set scope(value: string);
    private _authorizeTimeoutInSeconds?;
    get authorizeTimeoutInSeconds(): number;
    set authorizeTimeoutInSeconds(value: number);
    private _isAutoRefresh?;
    get isAutoRefresh(): boolean;
    set isAutoRefresh(value: boolean);
    /**
     * Build Laravel Passport Client based on given options.
     * @param options
     */
    constructor(options: LaravelPassportClientOptions);
    /**
     * ```js
     * lpClient.getToken();
     * ```
     *
     * Get the token this client has in cache.
     */
    getToken(): Promise<string | null>;
    /**
     * ```js
     * lpClient.getTokenScopes();
     * ```
     *
     * Get this client token's scope(s).
     */
    getTokenScopes(): string[] | null;
    /**
     * ```js
     * lpClient.getTokenExpiration();
     * ```
     *
     * Get this client token's expiration date.
     */
    getTokenExpiration(): Date | null;
    /**
     * ```js
     * lpClient.isTokenValid();
     * ```
     *
     * Returns `true` if the client has a token which is not expired, `false` otherwise.
     */
    isTokenValid(): boolean;
    /**
     * ```js
     * lpClient.getSignedInUserId();
     * ```
     *
     * Get this client token's user id.
     */
    getSignedInUserId(): number | null;
    /**
     * ```js
     * await lpClient.signIn();
     * ```
     *
     * Sign the client in. Starts with the iframe flow and fallsback to redirect flow if needed.
     * Resolves on `true` if the sign in has been successful, `false` otherwise.
     * @param {?string} scope
     */
    signIn(scope?: string): Promise<boolean>;
    /**
     * ```js
     * await lpClient.signInWithRedirect(scope);
     * ```
     *
     * Redirect to the authorize URL (`'/oauth/authorize'` by default) with appropriate
     * AuthorizeParameters. If provided, the given scope value will override the client's
     * default scope.
     * @param {?string} scope
     */
    signInWithRedirect(scope?: string): Promise<void>;
    /**
     * ```js
     * await lpClient.handleRedirectCallback();
     * ```
     *
     * Extract the authorization code returned in the query string, build the Authorization Signature and calls
     * `exchangeAuthorizationForToken` with it. Returns `true` if a token is obtained, `false`
     * otherwise.
     */
    handleRedirectCallback(): Promise<boolean>;
    /**
     * ```js
     * lpClient.signOut();
     * ```
     *
     * Remove the cached token.
     */
    signOut(): void;
    /**
     * Cache the token for this client.
     * @param token
     */
    private storeToken;
    /**
     * Build the authorization request URL for this client and given AuthorizeParameters.
     * @param authorizationRequest
     */
    private buildAuthorizeUrl;
    /**
     * Creates AuthorizationRequest for this client. If provided, the given scope value will override
     * the client's default scope.
     * @param scope
     */
    private buildAuthorizationRequest;
    /**
     * Execute the sign in process with given method for given scope. Resolves on `true` if the sign
     * in has been successful, `false` otherwise.
     * @param method
     * @param scope
     */
    private executeSignIn;
    /**
     * Use the given query string to make an Authorization and sign in with it.
     * @param queryString
     * @throws If given query string is empty
     */
    private convertToToken;
    /**
     * Parse the given query string and returns an Authorization with the parsed parameters.
     * @param queryString
     * @throws If given query string is empty
     */
    private makeAuthorization;
    /**
     * Exchange the given Authorization for a token. Returns `true` if sign in was successful,
     * `false` otherwise.
     * @param authorization
     * @throws if the recieved token's scope does not match the scope from authorization
     */
    private signInWithAuthorization;
    /**
     * Runs an XHTTP token request to the authentication server with given authorization signature.
     * @param authorizationSignature
     */
    private getTokenFromServer;
}
